# test SABCoM command line application 
import os
import sys

#myPath = os.path.dirname(os.path.abspath(__file__))
#sys.path.insert(0, myPath + '/../')


#from sabcom.__main__ import simulate, initialise


def test_initialise():
    pass
    #runner = cl.CliRunner()
    #default_initialisation_path = os.path.join(os.path.dirname(sys.path[0]), 'example_data/initialisations')
    #result = runner.invoke(initialise, [3, default_initialisation_path])
    #assert result.exit_code == 0
	
	
def test_initialise():
    pass
	#runner = cl.CliRunner()
    #default_output_path = os.path.join(os.path.dirname(sys.path[0]), 'example_data/output_data')
    #result = runner.invoke(simulate, [3, default_output_path])
    #assert result.exit_code == 0